/* scanargv - by Ronald D. Peters
Command line argument scanner with scanf-like syntax

Copyright 1991 by the Movies of the Future Group of the
MIT Media Laboratory.
All rights reserved.

Permission to use, copy, and/or modify this code and its documentation
for any purpose and without fee is hereby granted, provided that the
above copyright notice appear in supporting documentation, and that the
name of M.I.T. not be used in advertising or publicity pertaining to
distribution of the code without prior permission.  This code or any
other copies thereof may not be provided of otherwise made available to
others.  No title to or ownership of the code is hereby transferred.
M.I.T. makes no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.
*/

/*
Modifications to this file must be indicated in the following change log.
       $Log:	scanargv.c,v $
 * Revision 1.9  94/11/03  01:01:44  costa
 * Added support for GCC handling of standard arguments
 * 
 * Revision 1.8  1994/06/03  15:05:16  klee
 * changed <local/dat.h> to <dat/dat.h>
 *
 * Revision 1.8  1994/06/03  15:05:16  klee
 * changed <local/dat.h> to <dat/dat.h>
 *
 * Revision 1.7  1993/06/16  16:12:34  holtzman
 * Removed #include <malloc.h>
 *  -lacsap
 *
 * Revision 1.5  228/12/25  01:12:28  holtzman
 * Changed token from static char * to static const char *
 *
 * Renamed function strdup to strdpl to avoid name conflict with osf strdup.
 *
 * Revision 1.4  92/02/20  11:36:26  klee
 * Changed made by someone else.
 * 
 * Revision 1.3  91/11/02  22:35:11  klee
 * found lying around...
 * 
 * Version 1.0 10/30/91  yazoo@media-lab.media.mit.edu
 * Initial ANSI C version.
*
*/

/* #ifdef __GNUG__ */
#include <stdarg.h>
/* #else
   #include </usr/gnu/include/glibc/stdarg.h>
   #endif
 */ /* glibc/stdarg.h no longer exists !? dpwe 1994dec06 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
/* #include <malloc.h> OOPS */

/* Compilation options
   cc -DOPEN=d_open       %r, %w, %a opens DATFILE, not binary file
   cc -DOPEN=dxopen       open submatrix of DATFILE
                          otherwise OPEN is fopen
*/

#ifdef OPEN
#include <dat/dat.h>
#undef FILE
#define FILE DATFILE
#else
#define OPEN fopen
#endif

FILE *OPEN();

/* Suns don't have strtoul */
#ifdef sun
#define strtoul strtol
#endif

static const char *token;       /* Pointer to characters in format string */

typedef struct {          /* To get around a pointer to pointer to Node. */
  struct node *np;
} Node_point;

typedef struct node {
  char *tag;             /* The literal string from format line (eg, [], -o)*/
  struct node *parent;   /* The immediate parent of this object on the tree */
  Node_point *children;  /* Pointer to a list of node pointers to children */
  int num_child;         /* Number of children for this node */
  int type;              /* The type of argument associated with this node */
  int repeat;            /* Number of times its children may be repeated */
  unsigned int used;     /* Has this node been matched with any arguments? */
  void *variable;        /* Universal pointer to variables to link the */
                         /* conversions to. */
} Node;

typedef struct filelist {
  char *name;            /* The name of the file to be opened */
  FILE **dest;            /* The destination of the resulting FILE pointer */
  char *mode;            /* File mode (e.g., r, r+, w, a+, etc... */
} File_list;

static File_list files_to_open[512];  /* List of files to open after parse */
static int file_count = 0;           /* Total number of such files */

/* Function prototypes */

/* Utility functions */
static char *strdpl(char *);	 /* Duplicates a string */
static void *getarg(int, Node *); /* Grabs next variable off argument list */
static void skip_spaces(void);	 /* Skips spaces to next token */
static Node *nalloc(Node *,char *, int);       /* Allocates a new node */
/* Appends a node to a Node_point list */
static Node_point *append_node(Node_point *, Node *);
static Node *nth_child(Node *, int);   /* Returns nth child of a node */
static void reenable_args(Node *);     /* Used for repeatable expressions */
static int strtest(char *, char *);    /* Tests two strings, '*' is wildcard */

/* These functions take care of parsing the format string */
static Node *list(Node *);  /* Parses parens and []; definitions */
static Node *sub_list(Node *);   /* Takes care of the parts between the '|' */
static void repeat(Node *);      /* Parses ., .., ..., etc. definitions */
static Node *string(Node *);     /* Parses literal strings in format line */
static Node *conversion(Node *); /* Parses %foo definitions */
static Node_point *expression(Node *); /* Returns pointer to a list of Node
					 pointers of children */
/* These parse the command line input */
static int match_args(Node *, int, int, char **);
static int check_arg(char *);      /* Returns the type of argument */
static Node *olist_delve(Node *, int, char *); /* Finds a node of given type */
static Node *place_arg(Node *, char *);
static int open_files(void);

/* Enumerated names representing the type of variable va_arg points to */
enum {integer, floater, stringer, filer, increment, decrement, olist, sublist};

/* Global definition to simplify parameter passing */
va_list arg_pointer;

int scanargv(int argc, const char *const argv[], const char *usage,
	     const char *format, ...)
{
  Node *root;
  int i, result, ignore, flag = 0;
  char *ptr;
  
  token = format;
  files_to_open[0].name = NULL;
  file_count = 0;
  
  va_start(arg_pointer, format);
  root = nalloc(NULL, strdpl("home"), 0);
  ptr = (char *)format + strlen(format) - 1;
  if (*ptr == '\t')
    {
      ignore = 1;
      ptr = NULL;
    }
  else
    ignore = 0;
  if (*token == '*')    /* Ignore leading '*' if any to maintain */
    token++;            /* backward compatibility */
  if ((root->children = expression(root)) == NULL)
    {
      fprintf(stderr,"scanargv: Error in parse tree\n");
      exit(-1);
    }
  result = match_args(root, 1, argc, (char **)argv);
  if ((result == 1) && usage)
    {
      for (i = 0; i < root->num_child; i++)
	{
	  if ((nth_child(root, i))->type != olist)
	    flag = 1;
	}
      if (flag)
	{
	  fprintf(stderr, "\n");
	  fprintf(stderr, "usage: ");
	  fprintf(stderr, usage, argv[0]);
	  fprintf(stderr, "\n\n");
	  if (getenv("VERBOSE_SCANARGV"))
	    {
	      fprintf(stderr, "Current command line:\n");
	      for (i = 0; i < argc; i++)
		fprintf(stderr, "%s ", argv[i]);
	      fprintf(stderr, "\n\n");
	    }
	  exit(-1);
	}
    }
  else if ((result<argc && result>= -argc) && !ignore)
    {
      fprintf(stderr, "\n");
      fprintf(stderr, "usage: ");
      fprintf(stderr, usage, argv[0]);
      fprintf(stderr, "\n\n");
      if (getenv("VERBOSE_SCANARGV"))
	{
	  fprintf(stderr, "Current command line:\n");
	  for (i = 0; i < argc; i++)
	    fprintf(stderr, "%s ", argv[i]);
	  fprintf(stderr, "\n\n");
	}
      exit(-1);
    }
  if (open_files() == -1)
    exit(-1);
  return(result);
}


/* Duplicate a string to new memory */
static char *strdpl(char *s)
{
  char *p;

  p = (char *)malloc(strlen(s)+1);  /* +1 for '\0' */
  if (p != NULL)
    strcpy(p,s);
  return p;
}


/* Allocate and initialize a new node */
static Node *nalloc(Node *parent, char *tag, int type)
{
  Node *new_node;

  new_node = (Node *)malloc(sizeof(Node));
  new_node->parent = parent;
  new_node->tag = tag;
  new_node->children = NULL;
  new_node->num_child = 0;
  new_node->type = type;
  new_node->repeat = 0;
  new_node->used = 0;
  new_node->variable = NULL;
  return(new_node);
}

/* Skip white space */
static void skip_spaces(void)
{
  while (isspace(*token))
    token++;
}


/* Returns nth child of a node, 0 being the first child. */
static Node *nth_child(Node *in, int n)
{
  if (in != NULL)
    {
      if (n < in->num_child)
	return((in->children + n)->np);
    }
  return(NULL);   /* It shouldn't have gotten this far */
}


/* Appends a node to a node_pointer list */
static Node_point *append_node(Node_point *old_list, Node *new)
{
  Node_point *perm, *temp;
  int a, size = 0;

  if (old_list != NULL)     /* Find out size of list*/
    for (size = 0; (old_list + size)->np != NULL; size++);
  perm = (Node_point *)calloc(size + 2, sizeof(Node_point));
  for (a = 0; a < size; a++)
    {
      temp = perm + a;
      temp->np = (old_list + a)->np;
    }
  temp = perm + size;
  temp->np = new;    /* Add new node to expanded list */
  temp = perm + size + 1;
  temp->np = NULL;   /* Designates end of pointer list */
  if (old_list != NULL)
    free(old_list);
  return(perm);
}

#define __builtin_alignof sizeof

static void *getarg(int var_type, Node *parent)
{
  switch (var_type)
    {
    case floater:
      if (parent->tag[1] == 'l')
	/* return va_arg(arg_pointer, double *); */
        return ((double **)
		( arg_pointer  = (char *) 
		 ( (__builtin_alignof(double *) == 8) ? 
		   (((int) arg_pointer  + sizeof(double *) + 8 - 1) & -8) 
		 : (((int) arg_pointer  + sizeof(double *) + 4 - 1) & -4) ) 
		 )) [-1] ;
      else
	return va_arg(arg_pointer, float *);
      break;
    case increment:
    case decrement:
      return va_arg(arg_pointer, int *);
      break;
    case integer:
      switch (parent->tag[1])
	{
	case 'l':
	  return va_arg(arg_pointer, long int *);
	  break;
	case 'h':
	  return va_arg(arg_pointer, short int *);
	  break;
	default:
	  return va_arg(arg_pointer, int *);
	  break;
	}
    case stringer:
      return va_arg(arg_pointer, char *);
      break;
    case filer:
      return va_arg(arg_pointer, FILE *);
      break;
    default:
      return(NULL);     /* If it made it this far, something's wrong */
      break;
    }
}



static Node *conversion(Node *parent)
{
  char tempstring[4];
  char *s;
  Node *p;
  int var_type = 0;

  tempstring[0] = *token;     /* i.e., '%' conversion character */
  token++;
  s = tempstring + 1;

/* This automatically makes sure the grammar is correct as it parses */
  switch (*token)
    {
    case 'l':
    case 'h':
      *s++ = *token++;
      break;
    default:
      break;
    }
  switch (*token)     /* Numbers and strings */
    {
    case 'D':         /* For compatibility with pre-ANSI versions,     */
    case 'O':         /* scanargv supports capitalized modifiers which */
    case 'X':         /* have the same meaning as the 'l' modifier.    */
    case 'd':
    case 'o':
    case 'x':
    case 'u':
      var_type = integer;
      *s++ = *token++;
      break;
    case 'F':
    case 'f':
      var_type = floater;
      *s++ = *token++;
      break;
    case 's':
    case 'p':
    case 'c':
      var_type = stringer;
      *s++ = *token++;
      break;
    default:
      break;
    }
  switch(*token)   /* Files and inc/decrement operators */
    {
    case 'r':
    case 'w':
    case 'a':
      var_type = filer;
      *s++ = *token++;
      if (*token == '+')
	*s++ = *token++;
      break;
    case '+':
      var_type = increment;
      while(*token == '+')
	*s++ = *token++;
      break;
    case '-':
      var_type = decrement;
      while(*token == '-')
	*s++ = *token++;
      break;
    default:
      break;
    }
  *s = '\0';
  if (!strlen(tempstring))             /* Invalid conversion string */
    return(NULL);
  p = nalloc(parent, strdpl(tempstring), var_type);
  p->variable = getarg(var_type, p);
  return(p);
}



static Node *string(Node *parent)
{
  char tempstring[33];
  char *s;
  Node *p;

  s = tempstring;
  while(isalpha(*token) || *token == '-' || *token == '=' || *token == '_')
    {
      *s = *token;
      s++;
      token++;
    }
  if (*token == '*')
    *s++ = *token++;
  *s = '\0';
  p = nalloc(parent, strdpl(tempstring), stringer);
  skip_spaces();
  if ((p->children = expression(p)) != NULL) /* Error checking */
    return(p);
  else
    return(NULL);
}


static void repeat(Node *p)
{
  int i;

  for (i = 0; *token == '.'; i++)
    token++;
  p->repeat = i;
  return;
}


static Node *list(Node *parent)
{
  char tempstring[3];
  Node *p, *temp;
  Node_point *perm = NULL;

  tempstring[0] = *token;
  token++;
  p = nalloc(parent, NULL, olist);
  while(*token != ')' && *token != ']')
    {
      temp = sub_list(p);
      perm = append_node(perm , temp);
      p->num_child++;
    }
  p->children = perm;
  tempstring[1] = *token;
  token++;
  tempstring[2] = '\0';
  p->tag = strdpl(tempstring);
  switch (tempstring[0]) /* Make sure parentheses and braces match */
    {
    case '[':
      if (tempstring[1] != ']')
	return(NULL);
      break;
    case '(':
      if (tempstring[1] != ')')
	return(NULL);
      break;
    default:
      return(NULL);
      break;
    }
  if (p->children == NULL)
    return(NULL);
  else
    return(p);
}


static Node *sub_list(Node *parent)
{
  Node *p;

  p = nalloc(parent, strdpl("^"), sublist);
  /* '^' is the tag for each sub-list. i.e., the ^'s for [%d %f | %s]
     would be _%d %f_ and _%s_ */
  p->children = expression(p);
  if (p->children == NULL)
    return(NULL);
  else
    {
      if (*token == '|')
 	token++;         /* To skip the '|' element separator */
      return(p);
    }
}


static Node_point *expression(Node *parent)
{
  Node *temp;
  Node_point *perm = NULL;
  int i = 0, j = 0;

  while ((*token != '\0')&&(*token != '|')&&(*token != ']')&&(*token != ')'))
    {
      i++;
      skip_spaces();
      switch (*token)
 	{
 	case '%':
	  temp = conversion(parent);
	  break;
 	case '[':
	case '(':
	  temp = list(parent);
	  break;
 	case '.':
	  repeat(parent);
	  i--;
	  j = 1;
	  break;
 	default:
	  temp = string(parent);
	  break;
 	}
      if(temp == NULL)
	return(NULL);
      if (!j)
	perm = append_node(perm, temp);
      else
	j = 0;
      skip_spaces();
    }
  parent->num_child = i;
  return(perm);
}

static int match_args(Node *branch, int count, int argc, char **argv)
{
  int type, i=0, j, k=0, result;
  Node *temp = branch;

  if (temp == NULL)
    return(-count);
  if (argc == 1)
    return(1);
  while (temp != NULL)
    {
      if ((temp = nth_child(branch, i)) != NULL)
	{
	  for (j = 0, k = 0; (temp = nth_child(branch, i + j)) != NULL;
	       j++, k++)
	    {
	      if (temp->type == increment || temp->type == decrement)
		k--;
	      else if (argv[count + k] != NULL)
		{
		  if ((type = check_arg(argv[count + k])) != temp->type)
		    {
		      if (
			  ((temp->type == filer) && (type == stringer)) ||
			  ((temp->type == stringer) && (type == filer)) ||
			  (temp->type == stringer) ||
			  (temp->type == filer) ||
			  ((temp->type == floater) && (type == integer)))
			continue;
		      else
			break;
		    }
		}
	      else
		break;
	    }
	  if (temp == NULL || temp->type == olist)
	    {
	      for (k = 0; k < j; k++)
		{
		  temp = nth_child(branch, i++);
	      if (temp->type == increment || temp->type == decrement)
		place_arg(temp, NULL);
	      else
		place_arg(temp, argv[count++]);
		}
	    }
	}
      if (argv[count] == NULL)
	break;
      temp = branch;
      while (temp != NULL && argv[count] != NULL)
	{
	  if ((temp = olist_delve(branch, (type = check_arg(argv[count])),
				  argv[count])) != NULL)
	    {
	      if (temp->repeat > 0)
		{
		  for (k = 0; k < temp->repeat && argv[count] != NULL; k++)
		    {
		      j = 0;
		      if (temp->type == stringer &&
			  strtest(temp->tag, argv[count]))
			j = 1;
		      reenable_args(temp);
		      result = match_args(temp, count+j, argc, (char **)argv);
		      if (result < 0)
			break;
		      else
			count = result;
		    }
		  if (result < 0 && k == 0)
		    return(result);
		}
	      else if (temp->type == olist || temp->type == stringer)
		{
		  if (((k = strlen(temp->tag)) < strlen(argv[count])) &&
		      (temp->tag[k - 1] != '*'))
		    argv[count] += k;
		  else
		    count++;
		  count = match_args(temp, count, argc, (char **)argv);
		  if (count < 0)
		    return(count);
		  temp->used = 1;
		  temp->parent->used = 1;
		  if (!strcmp(temp->parent->parent->tag, "()"))
		    temp->parent->parent->used = 1;
		}
	      else
		{
		  count = match_args(temp, count, argc, (char **)argv);
		  if (count < 0)
		    return(count);
		  temp->used = 1;
		  temp->parent->used = 1;
		}
	    }
	}
      /* Skip to next block of arguments */
      while ((temp = nth_child(branch, i)) != NULL &&
	     temp->type != olist)
	i++;
      while ((temp = nth_child(branch, i)) != NULL &&
	     temp->type == olist)
	i++;
    }
  for (j = 0; (temp = nth_child(branch, j)) != NULL; j++)
    {       /* Make sure all mandatory arguments were used */
      switch (temp->type)
	{
	case olist:
	  break;
	default:
	  if (!temp->used)
	    return(-count);
	  break;
	}
    }
  return(count);
}


/* Allows the reuse of conversions in a repeatable expression */
static void reenable_args(Node *source)
{
  int i;
  Node *temp;

  for (i = 0; i < source->num_child; i++)
    {
      if ((temp = nth_child(source, i)) != NULL)
	if (temp->type != olist && temp->tag[1] != 's')
	  temp->used = 0;	  
    }
}


static int check_arg(char *arg)
{
  char *s;

  s = arg;

  while (isdigit(*s) || *s == '-' || *s == '+')
    {
      s++;
      if (*s == '\0')
	return(integer);
    }
  s = arg;
  while (isdigit(*s) || *s=='e' || *s=='E' || *s=='.' || *s=='+' || *s=='-')
    {
      s++;
      if (*s == '\0' && (s - arg) > 1)
	return(floater);
    }
  s = arg;
  if (*s++ == '0' && *s++ == 'x')
    {
      while (isxdigit(*s))
	{
	  s++;
	  if (*s == '\0')
	    return(integer);
	}
    }
  s = arg;
  while (isalpha(*s) || *s == '_' || *s == '*' || *s == '-' ||
	 *s == '=' || *s == '+' || *s == '^')
    {
      s++;
      if (*s == '\0' || *s == '/' || *s == '[' || *s == '.' ||
	  isdigit(*s) || *s == '^')
 	return(stringer);
    }
  if (*s == '.' || *s == '/')
    return(filer);
  return(-1);
}


/* Search the olists for a conversion of type */
static Node *olist_delve(Node *branch, int type, char *arg)
{
  int j = 0,
      k, leaf_type;

  Node *limb, *leaf, *temp;

  if (branch->parent == NULL)   /* If leaf is actually root */
    {
      while ((limb = nth_child(branch, j)) != NULL)
	{
	  if (limb->type == olist)  /* Find all olist children of root */
	    {
	      if ((leaf = olist_delve(limb, type, arg)) != NULL)
		return(leaf);
	    }
	  j++;
	}
      return(NULL);
    }
/* Branch is not root, so peruse limbs for this particular type of arg */
  for (j = 0; (limb = nth_child(branch, j)) != NULL; j++)
    {
      if (!limb->used)
	{
	  for (k = 0; (leaf = nth_child(limb, k)) != NULL; k++)
	    {
	      if (!leaf->used)
		{
		  switch (leaf->type)
		    {
		    case stringer:
		      if (strtest(leaf->tag, arg))
			return(leaf);
		      break;
		    case olist:
		      if ((temp = olist_delve(leaf, type, arg)) != NULL)
			return(temp);
		      break;
		    case sublist:
		    default:     /* %f %d %s, etc. */
		      if (leaf->children != NULL)
			{
			  if (((leaf_type=nth_child(leaf, 0)->type)==type) ||
			      (leaf_type == floater && type == integer) ||
			      (leaf_type == filer && type == stringer) ||
			      (leaf_type == stringer && type == filer))
			    return(leaf);
			}
		      if (((leaf->type)==type) ||
			  (leaf->type == floater && type == integer) ||
			  (leaf->type == filer && type == stringer) ||
			  (leaf_type == stringer && type == filer))
			return(limb);
		      break;
		    }
		}
	    }
	}
    }
  return(NULL);
}


static int strtest(char *tag, char *arg)
{
  while (*tag != '\0')
    {
      if (*tag != *arg)
	break;
      else
	{
	  tag++;
	  arg++;
	}
    }
  if ((*arg == '\0' && *tag == '\0')|| *tag == '*')
    return(1);
  else
    return(0);
}


static Node *place_arg(Node *source, char *arg)
{
  int k;
  float f;
  double d;
  Node *p = source->parent;
  static int j = 0;         /* Offset into array for repeatable expressions */
  static Node *previous = NULL; /* Last node modified */

  if (p != previous || p->parent == NULL)
    {
      j = 0;
      previous = p;
    }
  switch(source->type)
    {
    case integer:
      if (!strcmp(source->tag, "%D") || !strcmp(source->tag, "%ld"))
	{
	  *((long *)source->variable + j) = (long)atol(arg);
	  if (p->repeat > 0)
	    j++;
	}
      else if (!strcmp(source->tag, "%O") || !strcmp(source->tag, "%lo"))
	{
	  *((unsigned long *)source->variable + j) =
	    (unsigned long)strtoul(arg, NULL, 8);
	  if (p->repeat > 0)
	    j++;
	} 
      else if (!strcmp(source->tag, "%X") || !strcmp(source->tag, "%lx"))
	{
	  *((unsigned long *)source->variable + j) =
	    (unsigned long)strtoul(arg, NULL, 16);
	  if (p->repeat > 0)
	    j++;
	}
      else if (!strcmp(source->tag, "%lu"))
	{
	  *((unsigned long *)source->variable + j) =
	    (unsigned long)strtoul(arg, NULL, 16);
	  if (p->repeat > 0)
	    j++;
	}
      else if (source->tag[1] == 'h')
	{
	  if (source->tag[2] == 'x')
	    {
	      *((unsigned short *)source->variable + j) =
		(unsigned short)strtoul(arg, NULL, 16);
	      if (p->repeat > 0)
		j++;
	    }
	  else if (source->tag[2] == 'o')
	    {
	      *((unsigned short *)source->variable + j) =
		(unsigned short)strtoul(arg, NULL, 8);
	      if (p->repeat > 0)
		j++;
	    }
	  else if (source->tag[2] == 'u')
	    {
	      *((unsigned short *)source->variable + j) =
		(unsigned short)strtoul(arg, NULL, 8);
	      if (p->repeat > 0)
		j++;
	    }
	  else if (source->tag[1] == 'd' || source->tag[1] == 'i')
	    {
	      *((short *)source->variable + j) = (short)atol(arg);
	      if (p->repeat > 0)
		j++;
	    }
	  else
	    {
	      fprintf(stderr, "Conversion string %s not recognized\n",
		      source->tag);
	      return(NULL);
	    }
	}
      else
	{
	  if (source->tag[1] == 'x')
	    {
	      *((unsigned int *)source->variable + j) =
		(unsigned int)strtoul(arg, NULL, 16);
	      if (p->repeat > 0)
		j++;
	    }
	  else if (source->tag[1] == 'o')
	    {
	      *((unsigned int *)source->variable + j) =
		(unsigned int)strtoul(arg, NULL, 8);
	      if (p->repeat > 0)
		j++;
	    }
	  else if (source->tag[1] == 'u')
	    {
	      *((unsigned int *)source->variable + j) =
		(unsigned int)strtoul(arg, NULL, 8);
	      if (p->repeat > 0)
		j++;
	    }
	  else if (source->tag[1] == 'd' || source->tag[1] == 'i')
	    {
	      *((int *)source->variable + j) = (int)atol(arg);
	      if (p->repeat > 0)
		j++;
	    }
	  else
	    {
	      fprintf(stderr, "Conversion string %s not recognized\n",
		      source->tag);
	      return(NULL);
	    }
	}
      break;
    case floater:
      switch (source->tag[1])
	{
	case 'l':
	  sscanf(arg, "%lf", &d);
	  *((double *)source->variable + j) = d;
	  if (p->repeat > 0)
	    j++;
	  break;
	case 'f':
	case 'h':
	  sscanf(arg, "%f", &f);
	  *((float *)source->variable + j) = f;
	  if (p->repeat > 0)
	    j++;
	  break;
	default:
	  break;
	}
      break;
    case stringer:
      switch (source->tag[1])
	{
	case 'c':
	  *((char *)source->variable + j) = *arg;
	  if (p->repeat > 0)
	    j++;
	  break;
	case 'p':
	  *((char **)source->variable + j) = strdpl(arg);
	  if (p->repeat > 0)
	    j++;
	  break;
	case 's':  /* By nature, %s in not a repeatable expression */
	  for (k = 0; *arg != '\0'; k++, arg++)
	    *((char *)source->variable + k) = *arg;
	  *((char *)source->variable + k) = '\0';
	  break;
	default:
	  break;
	}
      break;
    case filer:
      files_to_open[file_count].name = strdpl(arg);
      files_to_open[file_count].dest = ((FILE **)source->variable + j);
      files_to_open[file_count].mode = strdpl(source->tag+1);
      files_to_open[file_count+1].name = NULL;
      file_count++;
      if (p->repeat > 0)
	j++;
      break;
    case increment:
      for (k=1; k < strlen(source->tag); k++)
	*((int *)source->variable) += 1;
      break;
    case decrement:
      for (j=1; j < strlen(source->tag); j++)
	*((int *)source->variable) -= 1;
      break;
    default:
      break;
    }
  source->used = 1;
  return(0);
}


static int open_files()
{
  int j;

  for (j = 0; j < file_count; j++)
    {
      switch (*files_to_open[j].mode)
	{
	case 'r':
	case 'a':
	  if ((*files_to_open[j].dest = OPEN(files_to_open[j].name,
		       files_to_open[j].mode)) == NULL)
	    j = -1;
	  break;
	default:    /* Skip w and w+ for now */
	  break;
	}
      if (j == -1)
	break;
    }
  if (j != -1)
    {
      for (j = 0; j < file_count; j++)
	{
	  switch (*files_to_open[j].mode)
	    {
	    case 'w':
	      if ((*files_to_open[j].dest = OPEN(files_to_open[j].name,
			   files_to_open[j].mode)) == NULL)
		j = -1;
	      break;
	    default:
	      break;
	    }
	  if (j == -1)
	    break;
	}
    }
  return(j);
}
